/**************************************
****   CDAUDIO.C  -  CD-DA Player  ****
**************************************/

//-- Include files -----------------------------------------------------------
#include <SDL.h>
#include <stdlib.h>
#include <string.h>
#include "../ym2610/adpcm.h"
#include "../ym2610/ym2610.h"

//-- Imported Variables ------------------------------------------------------
extern char	*global_error[80];

//-- Private Variables -------------------------------------------------------
static int			cdda_min_track;
static int			cdda_max_track;
static int			cdda_disk_length;
static int			cdda_track_start;
static int			cdda_track_end;
static int			cdda_loop_counter;
static SDL_CD		*cdrom;

//-- Public Variables --------------------------------------------------------
int			cdda_first_drive;
int			cdda_current_drive;
int			cdda_current_track;
int			cdda_current_frame;
int			cdda_playing;
int			cdda_autoloop;
int			cdda_volume;

//-- Function Prototypes -----------------------------------------------------
int			cdda_init(void);
int			cdda_play(int);
void		cdda_stop(void);
void		cdda_resume(void);
int			cdda_get_volume(void);
void		cdda_set_volume(int);
void		cdda_shutdown(void);
void		cdda_loop_check(void);
int 		cdda_get_disk_info(void);




//----------------------------------------------------------------------------
int	cdda_init(void)
{
	cdda_min_track = cdda_max_track = 0;
	cdda_current_track = 0;
	cdda_playing = 0;
	cdda_loop_counter = 0;

	/* Open the default drive */
	cdrom=SDL_CDOpen(cdda_current_drive);

	/* Did if open? Check if cdrom is NULL */
	if(!cdrom){
		fprintf(stderr, "Couldn't open drive: %s\n", SDL_GetError());
		exit(-1);
	}

	cdda_get_disk_info();
	return	1;
}

//----------------------------------------------------------------------------
int cdda_get_disk_info(void)
{
if(CD_INDRIVE(SDL_CDStatus(cdrom))) 
	{
		cdda_min_track = 0;
		cdda_max_track = cdrom->numtracks;
		cdda_disk_length = cdrom->numtracks;
		return 1;
	}
	else
	{
		printf("Error: No Disc in drive");
		exit(-1);
	}

}


//----------------------------------------------------------------------------
int	cdda_play(int track)
{
	if(CD_INDRIVE(SDL_CDStatus(cdrom))) 
	{
		SDL_CDPlayTracks(cdrom, track-1, 0, 1, 0);
		cdda_current_track = track;
		cdda_loop_counter=0;
		cdda_track_end=(cdrom->track[track-1].length*60)/CD_FPS;//Length in 1/60s of second
		cdda_playing = 1;
		return 1;
	}
	else
		return 0;
}

//----------------------------------------------------------------------------
void	cdda_stop(void)
{
	SDL_CDPause(cdrom);
	cdda_playing = 0;
}

//----------------------------------------------------------------------------
void	cdda_resume(void)
{
	SDL_CDResume(cdrom);	
	cdda_playing = 1;
}

//----------------------------------------------------------------------------
void	cdda_shutdown(void)
{
	SDL_CDStop(cdrom);
	SDL_CDClose(cdrom);
}

//----------------------------------------------------------------------------
void	cdda_loop_check(void)
{
	if (cdda_playing==1) {
		cdda_loop_counter++;
		if (cdda_loop_counter>=cdda_track_end) {
			if (cdda_autoloop)
				cdda_play(cdda_current_track);
			else
				cdda_stop();
		}
	}
}

//----------------------------------------------------------------------------
int	cdda_get_volume(void)
{
	return 0;
}

//----------------------------------------------------------------------------
void	cdda_set_volume(int volume)
{
}


