/**************************************
****  NEOCD.C  - Main Source File  ****
**************************************/

//-- Include Files -----------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <SDL.h>
#include <ctype.h>
#include "cdaudio/cdaudio.h"
#include "cdrom/cdrom.h"
#include "mc68000/68kregs.h"
#include "memory/memory.h"
#include "video/video.h"
#include "input/input.h"
#include "z80/z80intrf.h"
#include "sound/stream.h"
#include "ym2610/ym2610.h"
#include "ym2610/adpcm.h"
#include "icon.h"
#include <string.h>

extern char *strcat(char *_s1, const char *_s2);

// Version, date & time to display on startup
#define VERSION1 "NeoCD/SDL 0.1.8"
#define VERSION2 "Compiled on: "__DATE__" "__TIME__

char			neocd_wm_title[255]=VERSION1;

//-- Private variables -------------------------------------------------------
static int		cdda_was_playing = 0;

//-- Global Variables --------------------------------------------------------
char			*neogeo_rom_memory = NULL;
char			*neogeo_prg_memory = NULL;
char			*neogeo_fix_memory = NULL;
char			*neogeo_spr_memory = NULL;
char			*neogeo_pcm_memory = NULL;

char			global_error[80];
unsigned char	neogeo_memorycard[8192];
int				neogeo_prio_mode = 0;
int				neogeo_ipl_done = 0;
char			neogeo_region=1;
unsigned char	config_game_name[80];


//-- 68K Core related stuff --------------------------------------------------
int					mame_debug = 0;
char				*OP_ROM;
char				*OP_RAM;
int					previouspc = 0;
int					ophw = 0;
int					cur_mrhard = 0;
extern int			MC68000_ICount;
extern Registers68K	regs;
extern void			M68KRESET(void);	// 68K core init, not reset
extern int			M68KRUN(int);
extern void			initialize_memmap(void);

//-- Function Prototypes -----------------------------------------------------
void	neogeo_init(void);
void	neogeo_reset(void);
void	neogeo_hreset(void);
void	neogeo_shutdown(void);
void	MC68000_Cause_Interrupt(int);
void	neogeo_exception(void);
void	neogeo_run(void);
void	draw_main(void);
void	neogeo_quit(void);
void	not_implemented(void);
void	neogeo_machine_settings(void);
void	neogeo_debug_mode(void);
void	neogeo_cdda_check(void);
void	neogeo_cdda_control(void);
void	neogeo_do_cdda( int command, int trck_number_bcd);
void	neogeo_read_gamename(void);

//----------------------------------------------------------------------------
int	main(int argc, char* argv[])
{
	FILE			*fp;
	SDL_RWops	*iconfp;
	SDL_Surface *icon;
	int				result;
	//int				a, b, c,w,h,i;

	// Initialise SDL
	if((SDL_Init(SDL_INIT_VIDEO|SDL_INIT_CDROM|SDL_INIT_JOYSTICK)==-1)) { 
        printf("Could not initialize SDL: %s.\n", SDL_GetError());
        exit(-1);
    }


	//set up window icon
	iconfp=SDL_RWFromMem((void*)rawicon, rawiconsize);
	icon=SDL_LoadBMP_RW(iconfp,0);
	SDL_WM_SetIcon(icon,NULL);


	// Displays version number, date and time
	puts(VERSION1);
	puts(VERSION2);

	// Allocate needed memory buffers
	
	puts("NEOGEO: Allocating memory...");

	printf("PRG (2Mo) ... ");
	neogeo_prg_memory = calloc(1, 2097152);
	if (neogeo_prg_memory==NULL) {
		puts("failed !");
		return 0;
	}
	puts("OK!");

	printf("SPR (4Mo) ... ");
	neogeo_spr_memory = calloc(1, 4194304);
	if (neogeo_spr_memory==NULL) {
		puts("failed !");
		return 0;
	}
	puts("OK!");
	
	printf("ROM (128k) ... ");
	neogeo_rom_memory = calloc(1, 131072);
	if (neogeo_rom_memory==NULL) {
		puts("failed !");
		return 0;
	}
	puts("OK!");

	printf("FIX (128k) ... ");
	neogeo_fix_memory = calloc(1, 131072);
	if (neogeo_fix_memory==NULL) {
		puts("failed !");
		return 0;
	}
	puts("OK!");

	printf("PCM (1Mo) ... ");
	neogeo_pcm_memory = calloc(1, 1048576);
	if (neogeo_pcm_memory==NULL) {
		puts("failed !");
		return 0;
	}
	puts("OK!");
	
	// Initialize Memory Mapping
	initialize_memmap();

	// Read memory card
	fp = fopen("memcard.bin", "rb");
	if (fp!=NULL) {
		fread(neogeo_memorycard, 1, 8192, fp);
		fclose(fp);
	}

	// Register exit procedure
	atexit(neogeo_shutdown);
	
	// Video init
	result = video_init();
	if (!result)
	{
		video_mode = 0;
		puts(global_error);
		return 0;
	}

	//set window title
	SDL_WM_SetCaption(neocd_wm_title,neocd_wm_title);	


	//Hide Mouse Pointer
	SDL_ShowCursor(SDL_DISABLE);


	// Sound init
	YM2610Init();

	// Load BIOS
	fp = fopen("neocd.bin", "rb");
	if (fp==NULL) {
		printf("Fatal Error: Could not load NEOCD.BIN");
		return 0;
	}
	fread(neogeo_rom_memory, 1, 131072, fp);
	fclose(fp);
	swab(neogeo_rom_memory, neogeo_rom_memory, 131072);

	// Load startup RAM
	fp = fopen("startup.bin", "rb");
	if (fp==NULL) {
		printf("Fatal Error: Could not load STARTUP.BIN.");
		return 0;
	}
	fread(neogeo_prg_memory + 0x10F300, 1, 3328, fp);
	fclose(fp);
	swab(neogeo_prg_memory + 0x10F300, neogeo_prg_memory + 0x10F300, 3328);

	// Check BIOS validity
	if (*((short*)(neogeo_rom_memory+0xA822)) != 0x4BF9)
	{
		printf("Fatal Error: Invalid BIOS file.");
		return 0;
	}


	/*** Patch BIOS exit ***/
	*((short*)(neogeo_rom_memory+0x55e)) = 0xFABE;
	*((short*)(neogeo_rom_memory+0x560)) = 0x4E75;
	/*** Patch BIOS load files w/ now loading message ***/
	*((short*)(neogeo_rom_memory+0x552)) = 0xFABF;
	*((short*)(neogeo_rom_memory+0x554)) = 0x4E75;
	/*** Patch BIOS load files w/out now loading ***/
	*((short*)(neogeo_rom_memory+0x564)) = 0xFAC0;
	*((short*)(neogeo_rom_memory+0x566)) = 0x4E75;
	/*** Patch BIOS CDROM Check ***/
	*((short*)(neogeo_rom_memory+0xB040)) = 0x4E71;
	*((short*)(neogeo_rom_memory+0xB042)) = 0x4E71;
	/*** Patch BIOS upload command ***/
	*((short*)(neogeo_rom_memory+0x546)) = 0xFAC1;
	*((short*)(neogeo_rom_memory+0x548)) = 0x4E75;

	/*** Patch BIOS CDDA check ***/
	*((short*)(neogeo_rom_memory+0x56A)) = 0xFAC3;
	*((short*)(neogeo_rom_memory+0x56C)) = 0x4E75;

	/*** Full reset, please ***/
	*((short*)(neogeo_rom_memory+0xA87A)) = 0x4239;
	*((short*)(neogeo_rom_memory+0xA87C)) = 0x0010;
	*((short*)(neogeo_rom_memory+0xA87E)) = 0xFDAE;

	/*** Trap exceptions ***/
	*((short*)(neogeo_rom_memory+0xA5B6)) = 0x4AFC;

	
	// Install joystick(s)
	Input_Init_Joysticks();

	// Initialize CD-ROM
	cdrom_init();
	cdda_current_drive=cdrom_current_drive;
	cdda_init();

	// Initialize everything
	neogeo_init();
	neogeo_run();
	return 0;
}

//----------------------------------------------------------------------------
void	neogeo_init(void)
{
	M68KRESET();
}

//----------------------------------------------------------------------------
void	neogeo_hreset(void)
{

	FILE * fp;

	/* update window title with game name */
	neogeo_read_gamename();
	strcat(neocd_wm_title," - ");
	strcat(neocd_wm_title,config_game_name);
	SDL_WM_SetCaption(neocd_wm_title,neocd_wm_title);

	// Special patch
	if (strcmp(config_game_name, "TEST PROGRAM USA") == 0)
	{
		fp = fopen("patch.prg", "rb");
		fread(neogeo_prg_memory + 0x132000, 1, 112, fp);
		fclose(fp);
		swab(neogeo_prg_memory + 0x132000, neogeo_prg_memory + 0x132000, 112);
	}

	// First time init
	memset((void *)&regs, 0, sizeof(regs));
	regs.pc = 0xc0a822;
	regs.sr = 0x2700;
	regs.sr_h = 0x27;
	regs.a7 = regs.isp = regs.usp = 0x10F300;
	OP_ROM = neogeo_rom_memory - 0xC00000;
	MC68000_ICount = 0;
	cpu_writemem24_dword(0x10F6EE, cpu_readmem24_dword(0x68L)); // $68 *must* be copied at 10F6EE

	if (cpu_readmem24(0x107)&0x7E)
	{
		if (cpu_readmem24_word(0x13A))
		{
			cpu_writemem24_dword(0x10F6EA, (cpu_readmem24_word(0x13A)<<1) + 0xE00000);
		}
		else
		{
			cpu_writemem24_dword(0x10F6EA, 0);
			cpu_writemem24(0x00013B, 0x01);
		}
	}
	else
		cpu_writemem24_dword(0x10F6EA, 0xE1FDF0);

	/* Set System Region */
	cpu_writemem24(0x10FD83,neogeo_region);

	cdda_current_track = 0;
	cdda_get_disk_info();



	z80_init();
}	

//----------------------------------------------------------------------------
void	neogeo_reset(void)
{
	memset((void *)&regs, 0, sizeof(regs));
	regs.pc = 0x122;
	regs.sr = 0x2700;
	regs.sr_h = 0x27;
	regs.a7 = regs.isp = regs.usp = 0x10F300;
	OP_ROM = neogeo_prg_memory;
	MC68000_ICount = 0;
	cpu_writemem24(0x10FD80, 0x82);
	cpu_writemem24(0x10FDAF, 0x01);
	cpu_writemem24(0x10FEE1, 0x0A);
	cpu_writemem24(0x10F675, 0x01);
	cpu_writemem24(0x10FEBF, 0x00);
	cpu_writemem24_dword(0x10FDB6, 0);
	cpu_writemem24_dword(0x10FDBA, 0);

	/* System Region */
	cpu_writemem24(0x10FD83,neogeo_region);

	cdda_current_track = 0;
	z80_init();
}

//----------------------------------------------------------------------------
void	neogeo_shutdown(void)
{
	FILE	*fp;
	
	// Close everything and free memory

	if (sound_device != -1)
		YM2610Shutdown();

	cdda_shutdown();
	cdrom_shutdown();
	video_shutdown();

	SDL_Quit();

	puts("NEOGEO: System Shutdown.");

	fp = fopen("memcard.bin", "wb");
	fwrite(neogeo_memorycard, 1, 8192, fp);
	fclose(fp);

	free(neogeo_prg_memory);
	free(neogeo_rom_memory);
	free(neogeo_spr_memory);
	free(neogeo_fix_memory);
}

//----------------------------------------------------------------------------
void	neogeo_exception(void)
{
	// Users should never see this message !
	
	video_set_mode(VIDEO_TEXT);
	printf("NEOGEO: Exception Trapped at %08x !\n", previouspc);
	exit(0);
}	

//----------------------------------------------------------------------------
void MC68000_Cause_Interrupt(int level)
{
	if (level >= 1 && level <= 7)
        regs.irq |= 1 << (level-1);
}

//----------------------------------------------------------------------------
void	neogeo_exit(void)
{
	video_set_mode(VIDEO_TEXT);
	puts("NEOGEO: Exit requested by software...");
	exit(0);
}

//----------------------------------------------------------------------------
void	neogeo_run(void)
{
	int	i;
	int framecount=0;

	// If IPL.TXT not loaded, load it !
	if (!neogeo_ipl_done)
	{
		// Display Title
		cdrom_load_title();

		// Process IPL.TXT
		if (!cdrom_process_ipl()) {
			printf("Error: Error while processing IPL.TXT.");
			return;
		}
		
		// Reset everything
		neogeo_ipl_done = 1;
		neogeo_hreset();
	}
	
	// Resume cdda if it was playing
	if ((cdda_current_track != 0)&&(cdda_was_playing))
		cdda_resume();
		
	// Set sound volume
	//ASetAudioMixerValue(AUDIO_MIXER_MASTER_VOLUME, sound_vol);

	MC68000_ICount = 0;

	SDL_PollEvent(&event);
	key = SDL_GetKeyState(NULL);

	// Main loop
	while(!key[SDLK_ESCAPE])
	{
		// Execute Z80 timeslice (one VBL)
		mz80int(0);
		mz80exec(z80_cycles);
		z80_cycles = Z80_VBL_CYCLES;

		// One-vbl timeslice
		MC68000_ICount += 200000;
		M68KRUN(0);
		MC68000_Cause_Interrupt(2);
			
		// Check if there are pending commands for CDDA
		neogeo_cdda_check();
	
		if (sound_device != 0)
		{
			// Update audio channels
			for(i=0;i<6;i++)
				adpcm_ch_decodea(&adpcm_ch[i]);

			adpcm_ch_decodeb(&adpcm_ch[6]);
		}

		// Call display routine
		if(!frameskip)
			video_draw_screen1();
		else if(!(framecount++%(frameskip+1)))
			video_draw_screen1();

		cdda_loop_check();

		//Update keys and Joystick
		/*SDL_PollEvent(&event);
		key = SDL_GetKeyState(NULL);*/

		//process important events		
		if (event.type==SDL_QUIT)
			exit(0);
		if (key[SDLK_F12])
			video_save_snapshot();
		if (key[SDLK_F1])
			video_fullscreen_toggle();
		if (key[SDLK_F2])
			video_mode_toggle();
		if (key[SDLK_F3])
			incframeskip();
		if (key[SDLK_F4])
			sound_device=0;
	}

	// Save current state and return to menu
	cdda_was_playing = cdda_playing;

	// Clear audio buffers
	// sound_mute();

	// Stop CDDA
	if (cdda_playing)
		cdda_stop();
}

//----------------------------------------------------------------------------
// This is a really dirty hack to make SAMURAI SPIRITS RPG work
void	neogeo_prio_switch(void)
{
//	video_set_txt_mode();
//	debug_dump_regs();
//	clear_keybuf();
//	readkey();
//	video_set_gfx_mode();
	
	if (regs.d7 == 0xFFFF)
		return;
	
	if ((regs.d7==9)&&(regs.a3==0x10DED9)&&((regs.a2==0x1081d0)||((regs.a2&0xFFF000)==0x102000)))
	{
		neogeo_prio_mode = 0;
		return;
	}
	
	if ((regs.d7==8)&&(regs.a3==0x10DEC7)&&(regs.a2==0x102900))
	{
		neogeo_prio_mode = 0;
		return;
	}
	
	if (regs.a7 == 0x10F29C)
	{
		if ((regs.d4&0x4010)==0x4010)
		{
			neogeo_prio_mode = 0;
			return;
		}
		
		neogeo_prio_mode = 1;
	}
	else
	{
		if (regs.a3==0x5140)
		{
			neogeo_prio_mode = 1;
			return;
		}

		if ( (regs.a3&~0xF) == (regs.a4&~0xF) )
			neogeo_prio_mode = 1;
		else
			neogeo_prio_mode = 0;
	}
}

//----------------------------------------------------------------------------
void	not_implemented(void)
{
		printf("Error: This function isn't implemented.");
}

//----------------------------------------------------------------------------
void	neogeo_quit(void)
{
		exit(0);
}

//----------------------------------------------------------------------------
void neogeo_cdda_check(void)
{
	int		Offset;
	
	Offset = cpu_readmem24_dword(0x10F6EA);
	if (Offset < 0xE00000)	// Invalid addr
		return;

	Offset -= 0xE00000;
	Offset >>= 1;
	
	neogeo_do_cdda(subcpu_memspace[Offset], subcpu_memspace[Offset+1]);
}

//----------------------------------------------------------------------------
void neogeo_cdda_control(void)
{
	neogeo_do_cdda( (regs.d0>>8)&0xFF, regs.d0&0xFF );
}

//----------------------------------------------------------------------------
void neogeo_do_cdda( int command, int track_number_bcd)
{
	int		track_number;
	int		offset;

	if ((command == 0)&&(track_number_bcd == 0))
		return;

	cpu_writemem24(0x10F64B, track_number_bcd);
	cpu_writemem24(0x10F6F8, track_number_bcd);
	cpu_writemem24(0x10F6F7, command);
	cpu_writemem24(0x10F6F6, command);

	offset = cpu_readmem24_dword(0x10F6EA);

	if (offset)
	{
		offset -= 0xE00000;
		offset >>= 1;

		cpu_writemem24(0x10F678, 1);

		subcpu_memspace[offset] = 0;
		subcpu_memspace[offset+1] = 0;
	}

	switch( command )
	{
		case	0:
		case	1:
		case	5:
		case	4:
		case	3:
		case	7:
			track_number = ((track_number_bcd>>4)*10) + (track_number_bcd&0x0F);
			if ((track_number == 0)&&(!cdda_playing))
			{
				//sound_mute();
				cdda_resume();
			}
			else if ((track_number>1)&&(track_number<99))
			{
				//sound_mute();
				cdda_play(track_number);
				cdda_autoloop = !(command&1);
			}
			break;
		case	6:
		case	2:
			if (cdda_playing)
			{
				//sound_mute();
				cdda_stop();
			}
			break;
	}
}

//----------------------------------------------------------------------------
void neogeo_read_gamename(void)
{

	unsigned char	*Ptr;
	int				temp;

	Ptr = neogeo_prg_memory + cpu_readmem24_dword(0x11A);
	swab(Ptr, config_game_name, 80);
	
	for(temp=0;temp<80;temp++)
		if (!isprint(config_game_name[temp]))
			config_game_name[temp]=0;

}

