/************************
*** Z80 CPU Interface ***
************************/

//-- Include Files ----------------------------------------------------------
#include	"../ym2610/ym2610.h"
#include 	<stdio.h>
#include 	<stdlib.h>
#include	"mz80.h"

//-- Exported Functions -----------------------------------------------------
void	PortWrite(UINT16 PortNo, UINT8 Value, struct z80PortWrite *ptr);
UINT16	PortRead(UINT16 PortNo, struct z80PortRead *ptr);
UINT16 PortRead2(UINT16 PortNo, struct z80PortRead *ptr);

//-- Structures -------------------------------------------------------------
struct	z80PortRead ReadPorts[] =
{
	{0x0000,		0xffff,		PortRead,	NULL},
	{(UINT16)-1,	(UINT16)-1,	NULL,		NULL}
};

struct	z80PortWrite WritePorts[] =
{
	{0x0000,		0xffff,		PortWrite,	NULL},
	{(UINT16)-1,	(UINT16)-1,	NULL,		NULL}
};

struct MemoryReadByte	MemRead[] =
{
	{(UINT16)-1,	(UINT16)-1,	NULL,		NULL}
};

struct MemoryWriteByte	MemWrite[] =
{
	{(UINT16)-1,	(UINT16)-1,	NULL,		NULL}
};

//-- Variables --------------------------------------------------------------
CONTEXTMZ80	subcpu_context;
UINT8		subcpu_memspace[65536];
int			sound_code = 0;
int			pending_command = 0;
int			result_code = 0;
int			z80_cycles = Z80_VBL_CYCLES;

//---------------------------------------------------------------------------
void z80_init(void)
{
	subcpu_context.z80Base = subcpu_memspace;
	
	subcpu_context.z80IoRead  = ReadPorts;
	subcpu_context.z80IoWrite = WritePorts;
	
	subcpu_context.z80MemRead = MemRead;
	subcpu_context.z80MemWrite = MemWrite;
	
	mz80SetContext((void *)&subcpu_context);
	
	mz80reset();

	// Let Z80 do its initialization
	mz80exec(100000);
}

//---------------------------------------------------------------------------
void PortWrite(UINT16 PortNo, UINT8	Value, struct z80PortWrite *ptr)
{
	FILE	*fp;

/*	fp = fopen("C:\\TMP\\Z80.LOG", "at");
	fprintf(fp, "Write: %02X %02X\n", PortNo, Value);
	fclose(fp);*/

	switch( PortNo & 0xff)
	{
	case	0x4:
		YM2610Write(0, Value);
		break;
	
	case	0x5:
		YM2610Write(1, Value);
		break;
	
	case	0x6:
		YM2610Write(2, Value);
		break;
	
	case	0x7:
		YM2610Write(3, Value);
		break;

	case	0x8:
		/* NMI enable / acknowledge? (the data written doesn't matter) */
		break;
	
	case	0xC:
		result_code = Value;
		break;
	
	case	0x18:
		/* NMI disable? (the data written doesn't matter) */
		break;

	default:
		//printf("Unimplemented Z80 Write Port: %d\n",PortNo&0xff);
		break;
	}
}

//---------------------------------------------------------------------------
UINT16 PortRead(UINT16 PortNo, struct z80PortRead *ptr)
{
	int	Value;
	FILE	*fp;
	
	Value = PortRead2(PortNo, ptr);
	
/*	fp = fopen("C:\\TMP\\Z80.LOG", "at");
	fprintf(fp, "Read: %02X %02X\n", PortNo, Value);
	fclose(fp);*/
	
	return Value;
}

UINT16 PortRead2(UINT16 PortNo, struct z80PortRead *ptr)
{
	static int	coin_flip = 0;
	static int bank[4];
	
	switch( PortNo & 0xff)
	{
	case	0x0:
		pending_command = 0;
		return sound_code;
		break;
	
	case	0x4:
		coin_flip ^= 0x81;
//		return rand()&0x83;
		return YM2610Read(0);
		break;
	
	case	0x5:
		return YM2610Read(1);
		break;
	
	case	0x6:
		return YM2610Read(2);
		break;

	case 0x08:
		{
			bank[3] = 0x0800 * ((PortNo >> 8) & 0x7f);
			return 0;
			break;
		}
	case 0x09:
		{
			bank[2] = 0x1000 * ((PortNo >> 8) & 0x3f);
			break;
		}	
	case 0x0a:
		{
			bank[1] = 0x2000 * ((PortNo >> 8) & 0x1f);
			return 0;
			break;
		}
	case 0x0b:
		{
			bank[0] = 0x4000 * ((PortNo >> 8) & 0x0f);
			return 0;
			break;
		}
	default:
		//printf("Unimplemented Z80 Read Port: %d\n",PortNo&0xff);
		return 0;
		break;
	};	
	return 0;
}
